@@ -18,28 +18,32 @@ import ctypes
 import functools
 import inspect
 import socket
-import sys
+import time
 
 from oslo_log import log as logging
+import six
 
-from os_win._i18n import _LI
 from os_win import _utils
 from os_win import constants
 from os_win import exceptions
 from os_win.utils.storage import diskutils
-from os_win.utils.storage.initiator import iscsidsc_structures as iscsi_struct
-from os_win.utils.storage.initiator import iscsierr
 from os_win.utils import win32utils
+from os_win.utils.winapi import constants as w_const
+from os_win.utils.winapi.errmsg import iscsierr
+from os_win.utils.winapi import libs as w_lib
+from os_win.utils.winapi.libs import iscsidsc as iscsi_struct
 
-if sys.platform == 'win32':
-    iscsidsc = ctypes.windll.iscsidsc
+iscsidsc = w_lib.get_shared_lib_handle(w_lib.ISCSIDSC)
 
 LOG = logging.getLogger(__name__)
 
-ERROR_INSUFFICIENT_BUFFER = 0x7a
+
+def _get_buff(size, item_type):
+    buff = (ctypes.c_ubyte * size)()
+    return ctypes.cast(buff, ctypes.POINTER(item_type))
 
 
-def ensure_buff_and_retrieve_items(struct_type=None,
+def ensure_buff_and_retrieve_items(struct_type,
                                    func_requests_buff_sz=True,
                                    parse_output=True):
     # The iscsidsc.dll functions retrieving data accept a buffer, which will
@@ -47,12 +51,17 @@ def ensure_buff_and_retrieve_items(struct_type=None,
     # the error code will show it. In this case, the decorator will adjust the
     # buffer size based on the buffer size or the element count provided by
     # the function, attempting to call it again.
+    #
+    # We need to provide a buffer large enough to store the retrieved data.
+    # This may be more than the size of the retrieved structures as those
+    # may contain pointers. We're also casting the buffer to the appropriate
+    # type as function parameters are validated.
     def wrapper(f):
         @functools.wraps(f)
         def inner(*args, **kwargs):
             call_args = inspect.getcallargs(f, *args, **kwargs)
             call_args['element_count'] = ctypes.c_ulong(0)
-            call_args['buff'] = (ctypes.c_ubyte * 0)()
+            call_args['buff'] = _get_buff(0, struct_type)
             call_args['buff_size'] = ctypes.c_ulong(0)
 
             while True:
@@ -66,13 +75,14 @@ def ensure_buff_and_retrieve_items(struct_type=None,
                     else:
                         return ret_val
                 except exceptions.Win32Exception as ex:
-                    if (ex.error_code & 0xFFFF) == ERROR_INSUFFICIENT_BUFFER:
+                    if (ex.error_code & 0xFFFF ==
+                            w_const.ERROR_INSUFFICIENT_BUFFER):
                         if func_requests_buff_sz:
                             buff_size = call_args['buff_size'].value
                         else:
                             buff_size = (ctypes.sizeof(struct_type) *
                                          call_args['element_count'].value)
-                        call_args['buff'] = (ctypes.c_ubyte * buff_size)()
+                        call_args['buff'] = _get_buff(buff_size, struct_type)
                     else:
                         raise
         return inner
@@ -93,6 +103,7 @@ retry_decorator = functools.partial(
 class ISCSIInitiatorUtils(object):
     _DEFAULT_RESCAN_ATTEMPTS = 3
     _MS_IQN_PREFIX = "iqn.1991-05.com.microsoft"
+    _DEFAULT_ISCSI_PORT = 3260
 
     def __init__(self):
         self._win32utils = win32utils.Win32Utils()
@@ -110,7 +121,7 @@ class ISCSIInitiatorUtils(object):
         self._run_and_check_output(
             iscsidsc.ReportIScsiPersistentLoginsW,
             ctypes.byref(element_count),
-            ctypes.byref(buff),
+            buff,
             ctypes.byref(buff_size))
 
     @ensure_buff_and_retrieve_items(
@@ -124,19 +135,19 @@ class ISCSIInitiatorUtils(object):
             iscsidsc.ReportIScsiTargetsW,
             forced_update,
             ctypes.byref(element_count),
-            ctypes.byref(buff))
+            buff)
         return self._parse_string_list(buff, element_count.value)
 
     def get_iscsi_initiator(self):
         """Returns the initiator node name."""
         try:
-            buff = (ctypes.c_wchar * (iscsi_struct.MAX_ISCSI_NAME_LEN + 1))()
+            buff = (ctypes.c_wchar * (w_const.MAX_ISCSI_NAME_LEN + 1))()
             self._run_and_check_output(iscsidsc.GetIScsiInitiatorNodeNameW,
-                                       ctypes.byref(buff))
+                                       buff)
             return buff.value
         except exceptions.ISCSIInitiatorAPIException as ex:
-            LOG.info(_LI("The ISCSI initiator node name can't be found. "
-                         "Choosing the default one. Exception: %s"), ex)
+            LOG.info("The ISCSI initiator node name can't be found. "
+                     "Choosing the default one. Exception: %s", ex)
             return "%s:%s" % (self._MS_IQN_PREFIX, socket.getfqdn().lower())
 
     @ensure_buff_and_retrieve_items(
@@ -149,7 +160,7 @@ class ISCSIInitiatorUtils(object):
         self._run_and_check_output(
             iscsidsc.ReportIScsiInitiatorListW,
             ctypes.byref(element_count),
-            ctypes.byref(buff))
+            buff)
         return self._parse_string_list(buff, element_count.value)
 
     @staticmethod
@@ -160,6 +171,7 @@ class ISCSIInitiatorUtils(object):
         str_list = str_list.split('\x00') if str_list else []
         return str_list
 
+    @retry_decorator(error_codes=w_const.ERROR_INSUFFICIENT_BUFFER)
     def _login_iscsi_target(self, target_name, portal=None, login_opts=None,
                             is_persistent=True, initiator_name=None):
         session_id = iscsi_struct.ISCSI_UNIQUE_SESSION_ID()
@@ -176,10 +188,9 @@ class ISCSIInitiatorUtils(object):
             ctypes.c_wchar_p(target_name),
             False,  # IsInformationalSession
             initiator_name_ref,
-            ctypes.c_ulong(iscsi_struct.ISCSI_ANY_INITIATOR_PORT),
+            ctypes.c_ulong(w_const.ISCSI_ANY_INITIATOR_PORT),
             portal_ref,
-            iscsi_struct.ISCSI_SECURITY_FLAGS(
-                iscsi_struct.ISCSI_DEFAULT_SECURITY_FLAGS),
+            iscsi_struct.ISCSI_SECURITY_FLAGS(),
             None,  # Security flags / mappings (using default / auto)
             login_opts_ref,
             ctypes.c_ulong(0),
@@ -187,7 +198,7 @@ class ISCSIInitiatorUtils(object):
             is_persistent,
             ctypes.byref(session_id),
             ctypes.byref(connection_id),
-            ignored_error_codes=[iscsierr.ISDSC_TARGET_ALREADY_LOGGED_IN])
+            ignored_error_codes=[w_const.ISDSC_TARGET_ALREADY_LOGGED_IN])
         return session_id, connection_id
 
     @ensure_buff_and_retrieve_items(
@@ -198,15 +209,17 @@ class ISCSIInitiatorUtils(object):
             iscsidsc.GetIScsiSessionListW,
             ctypes.byref(buff_size),
             ctypes.byref(element_count),
-            ctypes.byref(buff))
+            buff)
 
     def _get_iscsi_target_sessions(self, target_name, connected_only=True):
         sessions = self._get_iscsi_sessions()
         return [session for session in sessions
-                if session.TargetNodeName == target_name and
-                (session.ConnectionCount > 0 or not connected_only)]
+                if session.TargetNodeName
+                and session.TargetNodeName.upper() == target_name.upper()
+                and (session.ConnectionCount > 0 or not connected_only)]
 
-    @retry_decorator(error_codes=iscsierr.ISDSC_SESSION_BUSY)
+    @retry_decorator(error_codes=(w_const.ISDSC_SESSION_BUSY,
+                                  w_const.ISDSC_DEVICE_BUSY_ON_SESSION))
     @ensure_buff_and_retrieve_items(
         struct_type=iscsi_struct.ISCSI_DEVICE_ON_SESSION,
         func_requests_buff_sz=False)
@@ -217,13 +230,13 @@ class ISCSIInitiatorUtils(object):
             iscsidsc.GetDevicesForIScsiSessionW,
             ctypes.byref(session_id),
             ctypes.byref(element_count),
-            ctypes.byref(buff))
+            buff)
 
     def _get_iscsi_session_disk_luns(self, session_id):
         devices = self._get_iscsi_session_devices(session_id)
         luns = [device.ScsiAddress.Lun for device in devices
                 if (device.StorageDeviceNumber.DeviceType ==
-                    iscsi_struct.FILE_DEVICE_DISK)]
+                    w_const.FILE_DEVICE_DISK)]
         return luns
 
     def _get_iscsi_device_from_session(self, session_id, target_lun):
@@ -232,26 +245,37 @@ class ISCSIInitiatorUtils(object):
             if device.ScsiAddress.Lun == target_lun:
                 return device
 
-    def _get_iscsi_device(self, target_name, target_lun):
-        sessions = self._get_iscsi_target_sessions(target_name)
-        for session in sessions:
-            device = self._get_iscsi_device_from_session(session.SessionId,
-                                                         target_lun)
-            if device:
-                return device
-
-    def get_device_number_for_target(self, target_name, target_lun):
+    def get_device_number_for_target(self, target_name, target_lun,
+                                     fail_if_not_found=False):
         # This method is preserved as it's used by the Hyper-V Nova driver.
-        device = self._get_iscsi_device(target_name, target_lun)
-        return device.StorageDeviceNumber.DeviceNumber if device else None
+        return self.get_device_number_and_path(target_name, target_lun,
+                                               fail_if_not_found)[0]
 
-    def get_device_number_and_path(self, target_name, target_lun):
+    def get_device_number_and_path(self, target_name, target_lun,
+                                   fail_if_not_found=False,
+                                   retry_attempts=10,
+                                   retry_interval=0.1,
+                                   rescan_disks=False,
+                                   ensure_mpio_claimed=False):
         # We try to avoid the need to seek the disk twice as this may take
         # unnecessary time.
-        device = self._get_iscsi_device(target_name, target_lun)
-        if device:
-            return device.StorageDeviceNumber.DeviceNumber, device.LegacyName
-        return None, None
+        device_number, device_path = None, None
+
+        try:
+            # Even if the disk was already discovered, under heavy load we may
+            # fail to locate it, in which case some retries will be performed.
+            (device_number,
+             device_path) = self.ensure_lun_available(
+                target_name, target_lun,
+                rescan_attempts=retry_attempts,
+                retry_interval=retry_interval,
+                rescan_disks=rescan_disks,
+                ensure_mpio_claimed=ensure_mpio_claimed)
+        except exceptions.ISCSILunNotAvailable:
+            if fail_if_not_found:
+                raise
+
+        return device_number, device_path
 
     def get_target_luns(self, target_name):
         # We only care about disk LUNs.
@@ -264,7 +288,7 @@ class ISCSIInitiatorUtils(object):
     def get_target_lun_count(self, target_name):
         return len(self.get_target_luns(target_name))
 
-    @retry_decorator(error_codes=iscsierr.ISDSC_SESSION_BUSY)
+    @retry_decorator(error_codes=w_const.ISDSC_SESSION_BUSY)
     def _logout_iscsi_target(self, session_id):
         self._run_and_check_output(
             iscsidsc.LogoutIScsiTarget,
@@ -281,21 +305,37 @@ class ISCSIInitiatorUtils(object):
                                    None)  # Portal group
 
     def _remove_static_target(self, target_name):
-        ignored_error_codes = [iscsierr.ISDSC_TARGET_NOT_FOUND]
+        ignored_error_codes = [w_const.ISDSC_TARGET_NOT_FOUND]
         self._run_and_check_output(iscsidsc.RemoveIScsiStaticTargetW,
                                    ctypes.c_wchar_p(target_name),
                                    ignored_error_codes=ignored_error_codes)
 
-    def _get_login_opts(self, auth_username, auth_password, auth_type,
-                        login_flags=0):
+    def _get_login_opts(self, auth_username=None, auth_password=None,
+                        auth_type=None, login_flags=0):
         if auth_type is None:
             auth_type = (constants.ISCSI_CHAP_AUTH_TYPE
                          if auth_username and auth_password
                          else constants.ISCSI_NO_AUTH_TYPE)
-        login_opts = iscsi_struct.ISCSI_LOGIN_OPTIONS(Username=auth_username,
-                                                      Password=auth_password,
-                                                      AuthType=auth_type,
-                                                      LoginFlags=login_flags)
+
+        login_opts = iscsi_struct.ISCSI_LOGIN_OPTIONS()
+
+        info_bitmap = 0
+        if auth_username:
+            login_opts.Username = six.b(auth_username)
+            login_opts.UsernameLength = len(auth_username)
+            info_bitmap |= w_const.ISCSI_LOGIN_OPTIONS_USERNAME
+
+        if auth_password:
+            login_opts.Password = six.b(auth_password)
+            login_opts.PasswordLength = len(auth_password)
+            info_bitmap |= w_const.ISCSI_LOGIN_OPTIONS_PASSWORD
+
+        login_opts.AuthType = auth_type
+        info_bitmap |= w_const.ISCSI_LOGIN_OPTIONS_AUTH_TYPE
+
+        login_opts.InformationSpecified = info_bitmap
+        login_opts.LoginFlags = login_flags
+
         return login_opts
 
     def _session_on_path_exists(self, target_sessions, portal_addr,
@@ -346,7 +386,7 @@ class ISCSIInitiatorUtils(object):
                              rescan_attempts=_DEFAULT_RESCAN_ATTEMPTS):
         portal_addr, portal_port = _utils.parse_server_string(target_portal)
         portal_port = (int(portal_port)
-                       if portal_port else iscsi_struct.DEFAULT_ISCSI_PORT)
+                       if portal_port else self._DEFAULT_ISCSI_PORT)
 
         known_targets = self.get_targets()
         if target_iqn not in known_targets:
@@ -360,9 +400,9 @@ class ISCSIInitiatorUtils(object):
             LOG.debug("Logging in iSCSI target %(target_iqn)s",
                       dict(target_iqn=target_iqn))
             # If the multipath flag is set, multiple sessions to the same
-            # target may be estabilished. MPIO must be enabled and configured
+            # target may be established. MPIO must be enabled and configured
             # to claim iSCSI disks, otherwise data corruption can occur.
-            login_flags = (iscsi_struct.ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED
+            login_flags = (w_const.ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED
                            if mpio_enabled else 0)
             login_opts = self._get_login_opts(auth_username,
                                               auth_password,
@@ -388,28 +428,53 @@ class ISCSIInitiatorUtils(object):
             self.ensure_lun_available(target_iqn, target_lun, rescan_attempts)
 
     def ensure_lun_available(self, target_iqn, target_lun,
-                             rescan_attempts=_DEFAULT_RESCAN_ATTEMPTS):
-        for attempt in range(rescan_attempts):
+                             rescan_attempts=_DEFAULT_RESCAN_ATTEMPTS,
+                             retry_interval=0,
+                             rescan_disks=True,
+                             ensure_mpio_claimed=False):
+        # This method should be called only after the iSCSI
+        # target has already been logged in.
+        for attempt in range(rescan_attempts + 1):
             sessions = self._get_iscsi_target_sessions(target_iqn)
             for session in sessions:
                 try:
                     sid = session.SessionId
                     device = self._get_iscsi_device_from_session(sid,
                                                                  target_lun)
-                    if device and (device.StorageDeviceNumber.DeviceNumber
-                                   not in (None, -1)):
-                        return
-                except exceptions.ISCSIInitiatorAPIException as ex:
-                    LOG.exception(ex)
+                    if not device:
+                        continue
+
+                    device_number = device.StorageDeviceNumber.DeviceNumber
+                    device_path = device.LegacyName
+
+                    if not device_path or device_number in (None, -1):
+                        continue
+
+                    if ensure_mpio_claimed and not (
+                            self._diskutils.is_mpio_disk(device_number)):
+                        LOG.debug("Disk %s was not claimed yet by the MPIO "
+                                  "service.", device_path)
+                        continue
+
+                    return device_number, device_path
+                except exceptions.ISCSIInitiatorAPIException:
+                    err_msg = ("Could not find lun %(target_lun)s  "
+                               "for iSCSI target %(target_iqn)s.")
+                    LOG.exception(err_msg,
+                                  dict(target_lun=target_lun,
+                                       target_iqn=target_iqn))
                     continue
             if attempt <= rescan_attempts:
-                self._diskutils.rescan_disks()
+                if retry_interval:
+                    time.sleep(retry_interval)
+                if rescan_disks:
+                    self._diskutils.rescan_disks()
 
         raise exceptions.ISCSILunNotAvailable(target_lun=target_lun,
                                               target_iqn=target_iqn)
 
-    @retry_decorator(error_codes=(iscsierr.ISDSC_SESSION_BUSY,
-                                  iscsierr.ISDSC_DEVICE_BUSY_ON_SESSION))
+    @retry_decorator(error_codes=(w_const.ISDSC_SESSION_BUSY,
+                                  w_const.ISDSC_DEVICE_BUSY_ON_SESSION))
     def logout_storage_target(self, target_iqn):
         LOG.debug("Logging out iSCSI target %(target_iqn)s",
                   dict(target_iqn=target_iqn))
